@@ -3,5 +3,5 @@
 // substitutes variables and creates bxversion.h.
 /////////////////////////////////////////////////////////////////////////
 
-#define VER_STRING "1.3"
-#define REL_STRING "December 10, 2001"
+#define VER_STRING "1.4"
+#define REL_STRING "March 27, 2002"
