@@ -1,10 +1,10 @@
 /////////////////////////////////////////////////////////////////////////
-// $Id: control.cc,v 1.40 2001/12/08 18:06:12 bdenney Exp $
+// $Id: control.cc,v 1.44 2002/03/26 14:46:02 bdenney Exp $
 /////////////////////////////////////////////////////////////////////////
 //
 /*
  * gui/control.cc
- * $Id: control.cc,v 1.40 2001/12/08 18:06:12 bdenney Exp $
+ * $Id: control.cc,v 1.44 2002/03/26 14:46:02 bdenney Exp $
  *
  * This is code for a text-mode control panel.  Note that this file
  * does NOT include bochs.h.  Instead, it does all of its contact with
@@ -74,6 +74,8 @@ extern "C" {
 
 #define CPANEL_PATH_LEN 512
 
+#define BX_INSERTED 11
+
 /* functions for changing particular options */
 void bx_control_panel_init ();
 int bx_read_rc (char *rc);
@@ -104,7 +106,7 @@ clean_string (char *s0)
 int 
 ask_uint (char *prompt, Bit32u min, Bit32u max, Bit32u the_default, Bit32u *out, int base)
 {
-  int n = max + 1;
+  Bit32u n = max + 1;
   char buffer[1024];
   char *clean;
   int illegal;
@@ -293,7 +295,7 @@ static char *startup_options_prompt =
 "4. Memory options\n"
 "5. Interface options\n"
 "6. Disk options\n"
-"7. Parallel port options\n"
+"7. Serial or Parallel port options\n"
 "8. Sound Blaster 16 options\n"
 "9. NE2000 network card options\n"
 "10. Other options\n"
@@ -338,9 +340,10 @@ static char *runtime_menu_prompt =
 "6. Log options for individual devices\n"
 "7. VGA Update Interval: %d\n"
 "8. Mouse: %s\n"
-"9. Instruction tracing: off (doesn't exist yet)\n"
-"10. Continue simulation\n"
-"11. Quit now\n"
+"9. Keyboard paste delay: %d\n"
+"10. Instruction tracing: off (doesn't exist yet)\n"
+"11. Continue simulation\n"
+"12. Quit now\n"
 "\n"
 "Please choose one:  [10] ";
 
@@ -372,18 +375,19 @@ void build_runtime_options_prompt (char *format, char *buf, int size)
   for (int i=0; i<2; i++) {
     SIM->get_floppy_options (i, &floppyop);
     sprintf (buffer[i], "%s, size=%s, %s", floppyop.Opath->getptr (),
-	  SIM->get_floppy_type_name (floppyop.Otype->get ()),
-	  floppyop.Oinitial_status->get () ? "inserted" : "ejected");
+      SIM->get_floppy_type_name (floppyop.Otype->get ()),
+      (floppyop.Oinitial_status->get () == BX_INSERTED)? "inserted" : "ejected");
     if (!floppyop.Opath->getptr ()[0]) strcpy (buffer[i], "none");
   }
   SIM->get_cdrom_options (0, &cdromop);
   sprintf (buffer[2], "%s, %spresent, %s",
-     cdromop.Opath->getptr (), cdromop.Opresent->get ()?"":"not ",
-     cdromop.Oinserted->get ()?"inserted":"ejected");
+    cdromop.Opath->getptr (), cdromop.Opresent->get ()?"":"not ",
+    (cdromop.Oinserted->get () == BX_INSERTED)? "inserted" : "ejected");
   snprintf (buf, size, format, buffer[0], buffer[1], buffer[2], 
       /* ips->get (), */
       SIM->get_param_num (BXP_VGA_UPDATE_INTERVAL)->get (), 
-      SIM->get_param_num (BXP_MOUSE_ENABLED)->get () ? "enabled" : "disabled");
+      SIM->get_param_num (BXP_MOUSE_ENABLED)->get () ? "enabled" : "disabled",
+      SIM->get_param_num (BXP_KBD_PASTE_DELAY)->get ());
 }
 
 int do_menu (bx_id id) {
@@ -451,7 +455,7 @@ int bx_control_panel (int menu)
 	 case 4: do_menu (BXP_MENU_MEMORY); break;
 	 case 5: do_menu (BXP_MENU_INTERFACE); break;
 	 case 6: do_menu (BXP_MENU_DISK); break;
-	 case 7: do_menu (BXP_MENU_PARALLEL); break;
+	 case 7: do_menu (BXP_MENU_SERIAL_PARALLEL); break;
 	 case 8: do_menu (BXP_SB16); break;
 	 case 9: do_menu (BXP_NE2K); break;
 	 case 10: do_menu (BXP_MENU_MISC); break;
@@ -462,7 +466,7 @@ int bx_control_panel (int menu)
    case BX_CPANEL_RUNTIME:
      char prompt[1024];
      build_runtime_options_prompt (runtime_menu_prompt, prompt, 1024);
-     if (ask_uint (prompt, 1, 11, 10, &choice, 10) < 0) return -1;
+     if (ask_uint (prompt, 1, 12, 11, &choice, 10) < 0) return -1;
      switch (choice) {
        case 1: do_menu (BXP_FLOPPYA); break;
        case 2: do_menu (BXP_FLOPPYB); break;
@@ -475,9 +479,10 @@ int bx_control_panel (int menu)
        case 6: bx_log_options (1); break;
        case 7: askparam (BXP_VGA_UPDATE_INTERVAL); break;
        case 8: askparam (BXP_MOUSE_ENABLED); break;
-       case 9: NOT_IMPLEMENTED (choice); break;
-       case 10: fprintf (stderr, "Continuing simulation\n"); return 0;
-       case 11:
+       case 9: askparam (BXP_KBD_PASTE_DELAY); break;
+       case 10: NOT_IMPLEMENTED (choice); break;
+       case 11: fprintf (stderr, "Continuing simulation\n"); return 0;
+       case 12:
 	 fprintf (stderr, "You chose quit on the control panel.\n");
 	 SIM->quit_sim (1);
 	 return -1;
@@ -796,7 +801,6 @@ bx_param_enum_c::text_ask (FILE *fpin, FILE *fpout)
   char *prompt = get_ask_format ();
   if (prompt == NULL) {
     // default prompt, if they didn't set an ask format string
-    char buffer[512];
     fprintf (fpout, "%s = ", get_name ());
     text_print (fpout);
     fprintf (fpout, "\n");
