@@ -1,8 +1,8 @@
 /////////////////////////////////////////////////////////////////////////
-// $Id: pic.cc,v 1.16 2001/11/27 18:15:39 bdenney Exp $
+// $Id: pic.cc,v 1.24 2002/03/25 01:31:59 bdenney Exp $
 /////////////////////////////////////////////////////////////////////////
 //
-//  Copyright (C) 2001  MandrakeSoft S.A.
+//  Copyright (C) 2002  MandrakeSoft S.A.
 //
 //    MandrakeSoft S.A.
 //    43, rue d'Aboukir
@@ -85,6 +85,9 @@ bx_pic_c::init(bx_devices_c *d)
   BX_PIC_THIS s.master_pic.init.requires_4 = 0;
   BX_PIC_THIS s.master_pic.init.byte_expected = 0;
   BX_PIC_THIS s.master_pic.special_mask = 0;
+  BX_PIC_THIS s.master_pic.lowest_priority = 7;
+  BX_PIC_THIS s.master_pic.polled = 0;
+  BX_PIC_THIS s.master_pic.rotate_on_autoeoi = 0;
 
   BX_PIC_THIS s.slave_pic.single_PIC = 0;
   BX_PIC_THIS s.slave_pic.interrupt_offset = 0x70; /* IRQ8 = INT 0x70 */
@@ -103,6 +106,14 @@ bx_pic_c::init(bx_devices_c *d)
   BX_PIC_THIS s.slave_pic.init.requires_4 = 0;
   BX_PIC_THIS s.slave_pic.init.byte_expected = 0;
   BX_PIC_THIS s.slave_pic.special_mask = 0;
+  BX_PIC_THIS s.slave_pic.lowest_priority = 7;
+  BX_PIC_THIS s.slave_pic.polled = 0;
+  BX_PIC_THIS s.slave_pic.rotate_on_autoeoi = 0;
+
+  for (unsigned i=0; i<8; i++) { /* all IRQ lines low */
+    BX_PIC_THIS s.master_pic.IRQ_line[i] = 0;
+    BX_PIC_THIS s.slave_pic.IRQ_line[i] = 0;
+  }
 }
 
 
@@ -137,6 +148,23 @@ bx_pic_c::read(Bit32u address, unsigned io_len)
    8259A PIC
    */
 
+  if((address == 0x20 || address == 0x21) && BX_PIC_THIS s.master_pic.polled) {
+    // In polled mode. Treat this as an interrupt acknowledge
+    clear_highest_interrupt(& BX_PIC_THIS s.master_pic);
+    BX_PIC_THIS s.master_pic.polled = 0;
+    service_master_pic();
+    return BX_PIC_THIS s.master_pic.irq;  // Return the current irq requested
+  }
+
+  if((address == 0xa0 || address == 0xa1) && BX_PIC_THIS s.slave_pic.polled) {
+    // In polled mode. Treat this as an interrupt acknowledge
+    clear_highest_interrupt(& BX_PIC_THIS s.slave_pic);
+    BX_PIC_THIS s.slave_pic.polled = 0;
+    service_slave_pic();
+    return BX_PIC_THIS s.slave_pic.irq;  // Return the current irq requested
+  }
+
+
   switch (address) {
     case 0x20:
       if (BX_PIC_THIS s.master_pic.read_reg_select) { /* ISR */
@@ -197,7 +225,6 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
 #else
   UNUSED(this_ptr);
 #endif // !BX_USE_PIC_SMF
-  int irq;
 
   if (io_len > 1)
     BX_PANIC(("io write to port %04x, len=%u", (unsigned) address,
@@ -217,10 +244,10 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
 	    // (mch) Ignore...
 	    // BX_INFO(("pic:master: init command 1 found %02x", (unsigned) value));
         if (bx_dbg.pic) {
-          BX_INFO(("pic:master: init command 1 found"));
-          BX_INFO(("            requires 4 = %u",
+          BX_INFO(("master: init command 1 found"));
+          BX_INFO(("        requires 4 = %u",
             (unsigned) (value & 0x01) ));
-          BX_INFO(("            cascade mode: [0=cascade,1=single] %u",
+          BX_INFO(("        cascade mode: [0=cascade,1=single] %u",
             (unsigned) ((value & 0x02) >> 1)));
           }
         BX_PIC_THIS s.master_pic.init.in_init = 1;
@@ -229,9 +256,18 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         BX_PIC_THIS s.master_pic.imr           = 0xFF; /* all IRQ's initially masked */
         BX_PIC_THIS s.master_pic.isr           = 0x00; /* no IRQ's in service */
         BX_PIC_THIS s.master_pic.irr           = 0x00; /* no IRQ's requested */
+        BX_PIC_THIS s.master_pic.lowest_priority = 7;
         BX_PIC_THIS s.master_pic.INT = 0; /* reprogramming clears previous INTR request */
+        BX_PIC_THIS s.master_pic.auto_eoi = 0;
+        BX_PIC_THIS s.master_pic.rotate_on_autoeoi = 0;
         if ( (value & 0x02) == 1 )
-          BX_PANIC(("pic:master: init command: single mode"));
+          BX_PANIC(("master: ICW1: single mode not supported"));
+        if ( (value & 0x08) == 1 ) {
+          BX_PANIC(("master: ICW1: level sensitive mode not supported"));
+	  }
+	else {
+          BX_DEBUG(("master: ICW1: edge triggered mode selected"));
+	  }
         BX_SET_INTR(0);
         return;
         }
@@ -242,8 +278,10 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         special_mask = (value & 0x60) >> 5;
         poll         = (value & 0x04) >> 2;
         read_op      = (value & 0x03);
-        if (poll)
-          BX_PANIC(("pic:master:OCW3: poll bit set"));
+        if (poll) {
+          BX_PIC_THIS s.master_pic.polled = 1;
+          return;
+        }
         if (read_op == 0x02) /* read IRR */
 	  BX_PIC_THIS s.master_pic.read_reg_select = 0;
         else if (read_op == 0x03) /* read ISR */
@@ -260,8 +298,10 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
 
       /* OCW2 */
       switch (value) {
-        case 0x00: // Rotate in Auto-EOI mode
-          BX_PANIC(("PIC: Rotate in Auto-EOI mode command received."));
+        case 0x00: // Rotate in auto eoi mode clear
+        case 0x80: // Rotate in auto eoi mode set
+          BX_PIC_THIS s.master_pic.rotate_on_autoeoi = (value != 0);
+          break;
 	case 0x0A: /* select read interrupt request register */
 	  BX_PIC_THIS s.master_pic.read_reg_select = 0;
 	  break;
@@ -269,16 +309,23 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
 	  BX_PIC_THIS s.master_pic.read_reg_select = 1;
 	  break;
 
-	case 0x20: /* end of interrupt command */
-          /* clear highest current in service bit */
-          for (irq=0; irq<=7; irq++) {
-            if (BX_PIC_THIS s.master_pic.isr & (1 << irq)) {
-              BX_PIC_THIS s.master_pic.isr &= ~(1 << irq);
-              break; /* out of for loop */
-              }
-            }
+        case 0xA0: // Rotate on non-specific end of interrupt
+        case 0x20: /* end of interrupt command */
+
+          clear_highest_interrupt(& BX_PIC_THIS s.master_pic);
+
+          if(value == 0xA0) {// Rotate in Auto-EOI mode
+            BX_PIC_THIS s.master_pic.lowest_priority ++;
+            if(BX_PIC_THIS s.master_pic.lowest_priority > 7)
+              BX_PIC_THIS s.master_pic.lowest_priority = 0;
+          }
+
           service_master_pic();
-	  break;
+          break;
+
+        case 0x40: // Intel PIC spec-sheet seems to indicate this should be ignored
+          BX_INFO(("IRQ no-op"));
+          break;
 
         case 0x60: /* specific EOI 0 */
         case 0x61: /* specific EOI 1 */
@@ -289,9 +336,8 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         case 0x66: /* specific EOI 6 */
         case 0x67: /* specific EOI 7 */
           BX_PIC_THIS s.master_pic.isr &= ~(1 << (value-0x60));
-          BX_PIC_THIS s.master_pic.irr &= ~(1 << (value-0x60));
           service_master_pic();
-	  break;
+          break;
 
         // IRQ lowest priority commands
         case 0xC0: // 0 7 6 5 4 3 2 1
@@ -302,12 +348,26 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         case 0xC5: // 5 4 3 2 1 0 7 6
         case 0xC6: // 6 5 4 3 2 1 0 7
         case 0xC7: // 7 6 5 4 3 2 1 0
-          // ignore for now
           BX_INFO(("IRQ lowest command 0x%x", value));
+          BX_PIC_THIS s.master_pic.lowest_priority = value - 0xC0;
+          break;
+
+        case 0xE0: // specific EOI and rotate 0
+        case 0xE1: // specific EOI and rotate 1
+        case 0xE2: // specific EOI and rotate 2
+        case 0xE3: // specific EOI and rotate 3
+        case 0xE4: // specific EOI and rotate 4
+        case 0xE5: // specific EOI and rotate 5
+        case 0xE6: // specific EOI and rotate 6
+        case 0xE7: // specific EOI and rotate 7
+          BX_PIC_THIS s.master_pic.isr &= ~(1 << (value-0xE0));
+          BX_PIC_THIS s.master_pic.lowest_priority = (value - 0xE0);
+          service_master_pic();
+
           break;
 
         default:
-          BX_PANIC(("PIC: write to port 20h = %02x", value));
+          BX_PANIC(("write to port 20h = %02x", value));
 	} /* switch (value) */
       break;
 
@@ -319,15 +379,15 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
             BX_PIC_THIS s.master_pic.interrupt_offset = value & 0xf8;
             BX_PIC_THIS s.master_pic.init.byte_expected = 3;
 	    if (bx_dbg.pic) {
-		  BX_INFO(("pic:master: init command 2 = %02x", (unsigned) value));
-		  BX_INFO(("            offset = INT %02x",
+		  BX_INFO(("master: init command 2 = %02x", (unsigned) value));
+		  BX_INFO(("        offset = INT %02x",
 			    BX_PIC_THIS s.master_pic.interrupt_offset));
 	    }
             return;
             break;
           case 3:
 	    if (bx_dbg.pic)
-		  BX_INFO(("pic:master: init command 3 = %02x", (unsigned) value));
+		  BX_INFO(("master: init command 3 = %02x", (unsigned) value));
             if (BX_PIC_THIS s.master_pic.init.requires_4) {
               BX_PIC_THIS s.master_pic.init.byte_expected = 4;
 	      }
@@ -337,11 +397,15 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
             return;
             break;
           case 4:
-	    if (bx_dbg.pic) {
-		  BX_INFO(("pic:master: init command 4 = %02x", (unsigned) value));
-		  if (value & 0x02) BX_INFO(("       auto EOI"));
-		  else BX_INFO(("normal EOI interrupt"));
-	    }
+            BX_DEBUG(("master: init command 4 = %02x", (unsigned) value));
+            if (value & 0x02) {
+              BX_DEBUG(("       auto EOI"));
+              BX_PIC_THIS s.master_pic.auto_eoi = 1;
+              }
+            else {
+              BX_DEBUG(("normal EOI interrupt"));
+              BX_PIC_THIS s.master_pic.auto_eoi = 0;
+              }
 	    if (value & 0x01) {
 		  if (bx_dbg.pic)
 			BX_INFO(("       80x86 mode"));
@@ -351,7 +415,7 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
             return;
             break;
           default:
-            BX_PANIC(("pic:master expecting bad init command"));
+            BX_PANIC(("master expecting bad init command"));
           }
         }
 
@@ -366,9 +430,9 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
     case 0xA0:
       if (value & 0x10) { /* initialization command 1 */
         BX_DEBUG(("slave: init command 1 found"));
-        BX_DEBUG(("           requires 4 = %u",
+        BX_DEBUG(("       requires 4 = %u",
             (unsigned) (value & 0x01) ));
-        BX_DEBUG(("           cascade mode: [0=cascade,1=single] %u",
+        BX_DEBUG(("       cascade mode: [0=cascade,1=single] %u",
             (unsigned) ((value & 0x02) >> 1)));
         BX_PIC_THIS s.slave_pic.init.in_init = 1;
         BX_PIC_THIS s.slave_pic.init.requires_4 = (value & 0x01);
@@ -376,9 +440,18 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         BX_PIC_THIS s.slave_pic.imr           = 0xFF; /* all IRQ's initially masked */
         BX_PIC_THIS s.slave_pic.isr           = 0x00; /* no IRQ's in service */
         BX_PIC_THIS s.slave_pic.irr           = 0x00; /* no IRQ's requested */
+        BX_PIC_THIS s.slave_pic.lowest_priority = 7;
         BX_PIC_THIS s.slave_pic.INT = 0; /* reprogramming clears previous INTR request */
+        BX_PIC_THIS s.slave_pic.auto_eoi = 0;
+        BX_PIC_THIS s.slave_pic.rotate_on_autoeoi = 0;
         if ( (value & 0x02) == 1 )
-          BX_PANIC(("slave: init command: single mode"));
+          BX_PANIC(("slave: ICW1: single mode not supported"));
+        if ( (value & 0x08) == 1 ) {
+          BX_PANIC(("slave: ICW1: level sensitive mode not supported"));
+	  }
+	else {
+          BX_DEBUG(("slave: ICW1: edge triggered mode selected"));
+	  }
         return;
         }
 
@@ -388,8 +461,10 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         special_mask = (value & 0x60) >> 5;
         poll         = (value & 0x04) >> 2;
         read_op      = (value & 0x03);
-        if (poll)
-          BX_PANIC(("slave:OCW3: poll bit set"));
+        if (poll) {
+          BX_PIC_THIS s.slave_pic.polled = 1;
+          return;
+        }
         if (read_op == 0x02) /* read IRR */
 	  BX_PIC_THIS s.slave_pic.read_reg_select = 0;
         else if (read_op == 0x03) /* read ISR */
@@ -405,22 +480,35 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         }
 
       switch (value) {
+        case 0x00: // Rotate in auto eoi mode clear
+        case 0x80: // Rotate in auto eoi mode set
+          BX_PIC_THIS s.slave_pic.rotate_on_autoeoi = (value != 0);
+          break;
+
 	case 0x0A: /* select read interrupt request register */
 	  BX_PIC_THIS s.slave_pic.read_reg_select = 0;
 	  break;
 	case 0x0B: /* select read interrupt in-service register */
 	  BX_PIC_THIS s.slave_pic.read_reg_select = 1;
 	  break;
-	case 0x20: /* end of interrupt command */
-          /* clear highest current in service bit */
-          for (irq=0; irq<=7; irq++) {
-            if (BX_PIC_THIS s.slave_pic.isr & (1 << irq)) {
-              BX_PIC_THIS s.slave_pic.isr &= ~(1 << irq);
-              break; /* out of for loop */
-              }
-            }
+
+        case 0xA0: // Rotate on non-specific end of interrupt
+        case 0x20: /* end of interrupt command */
+
+          clear_highest_interrupt(& BX_PIC_THIS s.slave_pic);
+
+          if(value == 0xA0) {// Rotate in Auto-EOI mode
+            BX_PIC_THIS s.slave_pic.lowest_priority ++;
+            if(BX_PIC_THIS s.slave_pic.lowest_priority > 7)
+              BX_PIC_THIS s.slave_pic.lowest_priority = 0;
+          }
+
           service_slave_pic();
-	  break;
+          break;
+
+        case 0x40: // Intel PIC spec-sheet seems to indicate this should be ignored
+          BX_INFO(("IRQ no-op"));
+          break;
 
         case 0x60: /* specific EOI 0 */
         case 0x61: /* specific EOI 1 */
@@ -431,9 +519,8 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         case 0x66: /* specific EOI 6 */
         case 0x67: /* specific EOI 7 */
           BX_PIC_THIS s.slave_pic.isr &= ~(1 << (value-0x60));
-          BX_PIC_THIS s.slave_pic.irr &= ~(1 << (value-0x60));
           service_slave_pic();
-	  break;
+          break;
 
         // IRQ lowest priority commands
         case 0xC0: // 0 7 6 5 4 3 2 1
@@ -444,12 +531,26 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
         case 0xC5: // 5 4 3 2 1 0 7 6
         case 0xC6: // 6 5 4 3 2 1 0 7
         case 0xC7: // 7 6 5 4 3 2 1 0
-          // ignore for now
           BX_INFO(("IRQ lowest command 0x%x", value));
+          BX_PIC_THIS s.slave_pic.lowest_priority = value - 0xC0;
+          break;
+
+        case 0xE0: // specific EOI and rotate 0
+        case 0xE1: // specific EOI and rotate 1
+        case 0xE2: // specific EOI and rotate 2
+        case 0xE3: // specific EOI and rotate 3
+        case 0xE4: // specific EOI and rotate 4
+        case 0xE5: // specific EOI and rotate 5
+        case 0xE6: // specific EOI and rotate 6
+        case 0xE7: // specific EOI and rotate 7
+          BX_PIC_THIS s.slave_pic.isr &= ~(1 << (value-0xE0));
+          BX_PIC_THIS s.slave_pic.lowest_priority = (value - 0xE0);
+          service_slave_pic();
+
           break;
 
         default:
-          BX_PANIC(("PIC: write to port A0h = %02x", value));
+          BX_PANIC(("write to port A0h = %02x", value));
 	} /* switch (value) */
       break;
 
@@ -462,13 +563,13 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
             BX_PIC_THIS s.slave_pic.init.byte_expected = 3;
 	    if (bx_dbg.pic) {
 		  BX_DEBUG(("slave: init command 2 = %02x", (unsigned) value));
-		  BX_DEBUG(("           offset = INT %02x",
+		  BX_DEBUG(("       offset = INT %02x",
 			    BX_PIC_THIS s.slave_pic.interrupt_offset));
 	    }
             return;
             break;
           case 3:
-		  BX_DEBUG(("slave: init command 3 = %02x", (unsigned) value));
+            BX_DEBUG(("slave: init command 3 = %02x", (unsigned) value));
             if (BX_PIC_THIS s.slave_pic.init.requires_4) {
               BX_PIC_THIS s.slave_pic.init.byte_expected = 4;
 	      	} else {
@@ -477,15 +578,20 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
             return;
             break;
           case 4:
-		if (bx_dbg.pic) {
-		      BX_DEBUG(("slave: init command 4 = %02x", (unsigned) value));
-		      if (value & 0x02) BX_INFO(("      auto EOI"));
-		      else BX_DEBUG(("normal EOI interrupt"));
-		}
-		if (value & 0x01) {
-		      if (bx_dbg.pic)
-			    BX_INFO(("      80x86 mode"));
-		} else BX_PANIC(("not 80x86 mode"));
+            BX_DEBUG(("slave: init command 4 = %02x", (unsigned) value));
+            if (value & 0x02) {
+              BX_DEBUG(("       auto EOI"));
+              BX_PIC_THIS s.slave_pic.auto_eoi = 1;
+              }
+            else {
+              BX_DEBUG(("normal EOI interrupt"));
+              BX_PIC_THIS s.slave_pic.auto_eoi = 0;
+              }
+	    if (value & 0x01) {
+		  if (bx_dbg.pic)
+			BX_INFO(("       80x86 mode"));
+	    } else
+		  BX_PANIC(("       not 80x86 mode"));
             BX_PIC_THIS s.slave_pic.init.in_init = 0;
             return;
             break;
@@ -506,6 +612,88 @@ bx_pic_c::write(Bit32u address, Bit32u value, unsigned io_len)
   return;
 }
 
+// new IRQ signal handling routines
+
+  void
+bx_pic_c::lower_irq(unsigned irq_no)
+{
+#if BX_SUPPORT_APIC
+  // forward this function call to the ioapic too
+  if ((BX_PIC_THIS devices) && (BX_PIC_THIS devices->ioapic))
+    BX_PIC_THIS devices->ioapic->untrigger_irq (irq_no, -1);
+#endif
+
+  if ((irq_no <= 7) && (BX_PIC_THIS s.master_pic.IRQ_line[irq_no])) {
+    BX_DEBUG(("IRQ line %d now low", (unsigned) irq_no));
+    BX_PIC_THIS s.master_pic.IRQ_line[irq_no] = 0;
+    BX_PIC_THIS s.master_pic.irr &= ~(1 << irq_no);
+    if ((BX_PIC_THIS s.master_pic.irr & ~BX_PIC_THIS s.master_pic.imr) == 0) {
+      BX_SET_INTR(0);
+      BX_PIC_THIS s.master_pic.INT = 0;
+    }
+  } else if ((irq_no > 7) && (irq_no <= 15) &&
+             (BX_PIC_THIS s.slave_pic.IRQ_line[irq_no-8])) {
+    BX_DEBUG(("IRQ line %d now low", (unsigned) irq_no));
+    BX_PIC_THIS s.slave_pic.IRQ_line[irq_no - 8] = 0;
+    BX_PIC_THIS s.slave_pic.irr &= ~(1 << (irq_no - 8));
+    if ((BX_PIC_THIS s.slave_pic.irr & ~BX_PIC_THIS s.slave_pic.imr) == 0) {
+      BX_PIC_THIS s.slave_pic.INT = 0;
+      lower_irq(2);
+    }
+  }
+}
+
+  void
+bx_pic_c::raise_irq(unsigned irq_no)
+{
+#if BX_SUPPORT_APIC
+  // forward this function call to the ioapic too
+  BX_PIC_THIS devices->ioapic->trigger_irq (irq_no, -1);
+#endif
+
+  if ((irq_no <= 7) && (!BX_PIC_THIS s.master_pic.IRQ_line[irq_no])) {
+    BX_DEBUG(("IRQ line %d now high", (unsigned) irq_no));
+    BX_PIC_THIS s.master_pic.IRQ_line[irq_no] = 1;
+    BX_PIC_THIS s.master_pic.irr |= (1 << irq_no);
+    service_master_pic();
+  } else if ((irq_no > 7) && (irq_no <= 15) &&
+             (!BX_PIC_THIS s.slave_pic.IRQ_line[irq_no-8])) {
+    BX_DEBUG(("IRQ line %d now high", (unsigned) irq_no));
+    BX_PIC_THIS s.slave_pic.IRQ_line[irq_no - 8] = 1;
+    BX_PIC_THIS s.slave_pic.irr |= (1 << (irq_no - 8));
+    service_slave_pic();
+  }
+}
+
+void  bx_pic_c::clear_highest_interrupt(bx_pic_t *pic)
+{
+  int irq;
+  int lowest_priority;
+  int highest_priority;
+
+  /* clear highest current in service bit */
+  lowest_priority = pic->lowest_priority;
+  highest_priority = lowest_priority + 1;
+  if(highest_priority > 7)
+    highest_priority = 0;
+
+  irq = highest_priority;
+  do {
+    if (pic->isr & (1 << irq)) {
+      pic->isr &= ~(1 << irq);
+      break; /* Return mask of bit cleared. */
+    }
+
+    irq ++;
+    if(irq > 7)
+      irq = 0;
+  } while(irq != highest_priority);
+
+}
+
+// old IRQ handling routines (disabled)
+
+#if 0
   void
 bx_pic_c::trigger_irq(unsigned irq_no)
 {
@@ -563,6 +751,7 @@ bx_pic_c::untrigger_irq(unsigned irq_no)
     }
   }
 }
+#endif
 
   /* */
   void
@@ -571,6 +760,9 @@ bx_pic_c::service_master_pic(void)
   Bit8u unmasked_requests;
   int irq;
   Bit8u isr, max_irq;
+  Bit8u highest_priority = BX_PIC_THIS s.master_pic.lowest_priority + 1;
+  if(highest_priority > 7)
+    highest_priority = 0;
 
   if (BX_PIC_THIS s.master_pic.INT) { /* last interrupt still not acknowleged */
     return;
@@ -580,28 +772,31 @@ bx_pic_c::service_master_pic(void)
     /* all priorities may be enabled.  check all IRR bits except ones
      * which have corresponding ISR bits set
      */
-    max_irq = 7;
+    max_irq = highest_priority;
     }
   else { /* normal mode */
     /* Find the highest priority IRQ that is enabled due to current ISR */
     isr = BX_PIC_THIS s.master_pic.isr;
     if (isr) {
-      max_irq = 0;
-      while ( (isr & 0x01) == 0 ) {
-        isr >>= 1;
+      max_irq = highest_priority;
+      while ( (isr & (1 << max_irq)) == 0) {
         max_irq++;
+        if(max_irq > 7)
+          max_irq = 0;
         }
-      if (max_irq == 0 ) return; /* IRQ0 in-service, no other priorities allowed */
+      if (max_irq == highest_priority ) return; /* Highest priority interrupt in-service,
+                                                 * no other priorities allowed */
       if (max_irq > 7) BX_PANIC(("error in service_master_pic()"));
       }
     else
-      max_irq = 7; /* 0..7 bits in ISR are cleared */
+      max_irq = highest_priority; /* 0..7 bits in ISR are cleared */
     }
 
 
   /* now, see if there are any higher priority requests */
   if ((unmasked_requests = (BX_PIC_THIS s.master_pic.irr & ~BX_PIC_THIS s.master_pic.imr)) ) {
-    for (irq=0; irq<=max_irq; irq++) {
+    irq = highest_priority;
+    do {
       /* for special mode, since we're looking at all IRQ's, skip if
        * current IRQ is already in-service
        */
@@ -609,14 +804,16 @@ bx_pic_c::service_master_pic(void)
         continue;
       if (unmasked_requests & (1 << irq)) {
         BX_DEBUG(("signalling IRQ(%u)", (unsigned) irq));
-        BX_PIC_THIS s.master_pic.irr &= ~(1 << irq);
-        /*??? do for slave too: BX_PIC_THIS s.master_pic.isr |=  (1 << irq);*/
         BX_PIC_THIS s.master_pic.INT = 1;
         BX_SET_INTR(1);
         BX_PIC_THIS s.master_pic.irq = irq;
         return;
         } /* if (unmasked_requests & ... */
-      } /* for (irq=7 ... */
+
+      irq ++;
+      if(irq > 7)
+        irq = 0;
+      } while(irq != max_irq); /* do ... */
     } /* if (unmasked_requests = ... */
 }
 
@@ -626,54 +823,62 @@ bx_pic_c::service_slave_pic(void)
 {
   Bit8u unmasked_requests;
   int irq;
-  Bit8u isr, lowest_priority_irq;
+  Bit8u isr, max_irq;
+  Bit8u highest_priority = BX_PIC_THIS s.slave_pic.lowest_priority + 1;
+  if(highest_priority > 7)
+    highest_priority = 0;
 
   if (BX_PIC_THIS s.slave_pic.INT) { /* last interrupt still not acknowleged */
     return;
     }
 
-  if (BX_PIC_THIS s.master_pic.special_mask) {
+  if (BX_PIC_THIS s.slave_pic.special_mask) {
     /* all priorities may be enabled.  check all IRR bits except ones
      * which have corresponding ISR bits set
      */
-    lowest_priority_irq = 8;
-    }
-  else {
-  /* Find the highest priority IRQ that is enabled due to current ISR */
-  isr = BX_PIC_THIS s.slave_pic.isr;
-  if (isr) {
-    lowest_priority_irq = 0;
-    while ( !(isr & 0x01) ) {
-      isr >>= 1;
-      lowest_priority_irq++;
+    max_irq = highest_priority;
+  }
+  else { /* normal mode */
+    /* Find the highest priority IRQ that is enabled due to current ISR */
+    isr = BX_PIC_THIS s.slave_pic.isr;
+    if (isr) {
+      max_irq = highest_priority;
+      while ( (isr & (1 << max_irq)) == 0) {
+        max_irq++;
+        if(max_irq > 7)
+          max_irq = 0;
+        }
+      if (max_irq == highest_priority ) return; /* Highest priority interrupt in-service,
+                                                 * no other priorities allowed */
+      if (max_irq > 7) BX_PANIC(("error in service_master_pic()"));
       }
-    if (lowest_priority_irq > 7) BX_PANIC(("error in service_slave_pic()"));
-    }
-  else
-    lowest_priority_irq = 8;
+    else
+      max_irq = highest_priority; /* 0..7 bits in ISR are cleared */
   }
 
 
   /* now, see if there are any higher priority requests */
   if ((unmasked_requests = (BX_PIC_THIS s.slave_pic.irr & ~BX_PIC_THIS s.slave_pic.imr)) ) {
-    for (irq=0; irq<lowest_priority_irq; irq++) {
+    irq = highest_priority;
+    do {
       /* for special mode, since we're looking at all IRQ's, skip if
        * current IRQ is already in-service
        */
       if ( BX_PIC_THIS s.slave_pic.special_mask && ((BX_PIC_THIS s.slave_pic.isr >> irq) & 0x01) )
         continue;
       if (unmasked_requests & (1 << irq)) {
-        if (bx_dbg.pic)
-          BX_DEBUG(("slave: signalling IRQ(%u)",
-            (unsigned) 8 + irq));
-        BX_PIC_THIS s.slave_pic.irr &= ~(1 << irq);
+        BX_DEBUG(("slave: signalling IRQ(%u)", (unsigned) 8 + irq));
+
         BX_PIC_THIS s.slave_pic.INT = 1;
-        BX_PIC_THIS s.master_pic.irr |= 0x04; /* request IRQ 2 on master pic */
         BX_PIC_THIS s.slave_pic.irq = irq;
-        service_master_pic();
+        raise_irq(2); /* request IRQ 2 on master pic */
         return;
         } /* if (unmasked_requests & ... */
-      } /* for (irq=7 ... */
+
+        irq ++;
+        if(irq > 7)
+          irq = 0;
+      } while(irq != max_irq); /* do ... */
     } /* if (unmasked_requests = ... */
 }
 
@@ -687,8 +892,12 @@ bx_pic_c::IAC(void)
 
   BX_SET_INTR(0);
   BX_PIC_THIS s.master_pic.INT = 0;
-  BX_PIC_THIS s.master_pic.isr |= (1 << BX_PIC_THIS s.master_pic.irq);
   BX_PIC_THIS s.master_pic.irr &= ~(1 << BX_PIC_THIS s.master_pic.irq);
+  // In autoeoi mode don't set the isr bit.
+  if(!BX_PIC_THIS s.master_pic.auto_eoi)
+    BX_PIC_THIS s.master_pic.isr |= (1 << BX_PIC_THIS s.master_pic.irq);
+  else if(BX_PIC_THIS s.slave_pic.rotate_on_autoeoi)
+    BX_PIC_THIS s.slave_pic.lowest_priority = BX_PIC_THIS s.master_pic.irq;
 
   if (BX_PIC_THIS s.master_pic.irq != 2) {
     irq    = BX_PIC_THIS s.master_pic.irq;
@@ -696,10 +905,15 @@ bx_pic_c::IAC(void)
     }
   else { /* IRQ2 = slave pic IRQ8..15 */
     BX_PIC_THIS s.slave_pic.INT = 0;
+    BX_PIC_THIS s.master_pic.IRQ_line[2] = 0;
     irq    = BX_PIC_THIS s.slave_pic.irq;
     vector = irq + BX_PIC_THIS s.slave_pic.interrupt_offset;
-    BX_PIC_THIS s.slave_pic.isr |= (1 << BX_PIC_THIS s.slave_pic.irq);
     BX_PIC_THIS s.slave_pic.irr &= ~(1 << BX_PIC_THIS s.slave_pic.irq);
+    // In autoeoi mode don't set the isr bit.
+    if(!BX_PIC_THIS s.slave_pic.auto_eoi)
+      BX_PIC_THIS s.slave_pic.isr |= (1 << BX_PIC_THIS s.slave_pic.irq);
+    else if(BX_PIC_THIS s.slave_pic.rotate_on_autoeoi)
+      BX_PIC_THIS s.slave_pic.lowest_priority = BX_PIC_THIS s.slave_pic.irq;
     service_slave_pic();
     irq += 8; // for debug printing purposes
     }
